@@ -20,6 +20,10 @@
 #include <mysys_err.h>
 #include <my_getopt.h>
 #include <errno.h>
+#include <m_string.h>
+
+typedef void (*init_func_p)(const struct my_option *option, uchar* *variable,
+                            longlong value);
 
 static void default_reporter(enum loglevel level, const char *format, ...);
 my_error_reporter my_getopt_error_reporter= &default_reporter;
@@ -31,11 +35,16 @@ my_bool getopt_compare_strings(const char *s,
 			       const char *t,
 			       uint length);
 static longlong getopt_ll(char *arg, const struct my_option *optp, int *err);
-static longlong getopt_ll_limit_value(longlong, const struct my_option *);
 static ulonglong getopt_ull(char *arg, const struct my_option *optp,
 			    int *err);
-static void init_variables(const struct my_option *options);
-static int setval(const struct my_option *opts, gptr *value, char *argument,
+static double getopt_double(char *arg, const struct my_option *optp, int *err);
+static void init_variables(const struct my_option *options,
+                           init_func_p init_one_value);
+static void init_one_value(const struct my_option *option, uchar* *variable,
+			   longlong value);
+static void fini_one_value(const struct my_option *option, uchar* *variable,
+			   longlong value);
+static int setval(const struct my_option *opts, uchar* *value, char *argument,
 		  my_bool set_maximum_value);
 static char *check_struct_option(char *cur_arg, char *key_name);
 
@@ -60,6 +69,13 @@ char *disabled_my_option= (char*) "0";
 
 my_bool my_getopt_print_errors= 1;
 
+/* 
+   This is a flag that can be set in client programs. 1 means that
+   my_getopt will skip over options it does not know how to handle.
+*/
+
+my_bool my_getopt_skip_unknown= 0;
+
 static void default_reporter(enum loglevel level,
                              const char *format, ...)
 {
@@ -85,10 +101,10 @@ static void default_reporter(enum loglevel level,
   one. Call function 'get_one_option()' once for each option.
 */
 
-static gptr* (*getopt_get_addr)(const char *, uint, const struct my_option *);
+static uchar** (*getopt_get_addr)(const char *, uint, const struct my_option *, int *);
 
-void my_getopt_register_get_addr(gptr* (*func_addr)(const char *, uint,
-						    const struct my_option *))
+void my_getopt_register_get_addr(uchar** (*func_addr)(const char *, uint,
+						    const struct my_option *, int *))
 {
   getopt_get_addr= func_addr;
 }
@@ -97,22 +113,26 @@ int handle_options(int *argc, char ***argv,
 		   const struct my_option *longopts,
                    my_get_one_option get_one_option)
 {
-  uint opt_found, argvpos= 0, length, i;
+  uint opt_found, argvpos= 0, length;
   my_bool end_of_options= 0, must_be_var, set_maximum_value,
           option_is_loose;
-  char **pos, **pos_end, *optend, *prev_found,
+  char **pos, **pos_end, *optend, *UNINIT_VAR(prev_found),
        *opt_str, key_name[FN_REFLEN];
   const struct my_option *optp;
-  gptr *value;
-  int error;
+  uchar* *value;
+  int error, i;
 
   LINT_INIT(opt_found);
+  /* handle_options() assumes arg0 (program name) always exists */
+  DBUG_ASSERT(argc && *argc >= 1);
+  DBUG_ASSERT(argv && *argv);
   (*argc)--; /* Skip the program name */
   (*argv)++; /*      --- || ----      */
-  init_variables(longopts);
+  init_variables(longopts, init_one_value);
 
   for (pos= *argv, pos_end=pos+ *argc; pos != pos_end ; pos++)
   {
+    char **first= pos;
     char *cur_arg= *pos;
     if (cur_arg[0] == '-' && cur_arg[1] && !end_of_options) /* must be opt */
     {
@@ -219,12 +239,11 @@ int handle_options(int *argc, char ***argv,
 		/*
 		  We were called with a special prefix, we can reuse opt_found
 		*/
-		opt_str+= (special_opt_prefix_lengths[i] + 1);
+		opt_str+= special_opt_prefix_lengths[i] + 1;
+                length-= special_opt_prefix_lengths[i] + 1;
 		if (i == OPT_LOOSE)
 		  option_is_loose= 1;
-		if ((opt_found= findopt(opt_str, length -
-					(special_opt_prefix_lengths[i] + 1),
-					&optp, &prev_found)))
+		if ((opt_found= findopt(opt_str, length, &optp, &prev_found)))
 		{
 		  if (opt_found > 1)
 		  {
@@ -248,7 +267,7 @@ int handle_options(int *argc, char ***argv,
 		    break;
 		  case OPT_ENABLE:
 		    optend= (optend && *optend == '0' && !(*(optend + 1))) ?
- 		      disabled_my_option : (char*) "1";
+                      disabled_my_option : (char*) "1";
 		    break;
 		  case OPT_MAXIMUM:
 		    set_maximum_value= 1;
@@ -257,11 +276,25 @@ int handle_options(int *argc, char ***argv,
 		  }
 		  break; /* break from the inner loop, main loop continues */
 		}
+                i= -1; /* restart the loop */
 	      }
 	    }
 	  }
 	  if (!opt_found)
 	  {
+            if (my_getopt_skip_unknown)
+            {
+              /*
+                preserve all the components of this unknown option, this may
+                occurr when the user provides options like: "-O foo" or
+                "--set-variable foo" (note that theres a space in there)
+                Generally, these kind of options are to be avoided
+              */
+              do {
+                (*argv)[argvpos++]= *first++;
+              } while (first <= pos);
+              continue;
+            }
 	    if (must_be_var)
 	    {
 	      if (my_getopt_print_errors)
@@ -275,9 +308,9 @@ int handle_options(int *argc, char ***argv,
 	    else
 	    {
 	      if (my_getopt_print_errors)
-                my_getopt_error_reporter(option_is_loose ?
+                my_getopt_error_reporter(option_is_loose ? 
                                            WARNING_LEVEL : ERROR_LEVEL,
-                                         "%s: unknown option '--%s'",
+                                         "%s: unknown option '--%s'", 
                                          my_progname, cur_arg);
 	      if (!option_is_loose)
 		return EXIT_UNKNOWN_OPTION;
@@ -330,8 +363,12 @@ int handle_options(int *argc, char ***argv,
                                      my_progname, optp->name);
 	  return EXIT_NO_ARGUMENT_ALLOWED;
 	}
+        error= 0;
 	value= optp->var_type & GET_ASK_ADDR ?
-	  (*getopt_get_addr)(key_name, (uint) strlen(key_name), optp) : optp->value;
+	  (*getopt_get_addr)(key_name, (uint) strlen(key_name), optp, &error) :
+          optp->value;
+        if (error)
+          return error;
   
 	if (optp->arg_type == NO_ARG)
 	{
@@ -365,15 +402,17 @@ invalid value '%s'",
 				       my_progname, optp->name, optend);
 	      continue;
 	    }
-	    get_one_option(optp->id, optp,
-			   *((my_bool*) value) ?
-			   (char*) "1" : disabled_my_option);
+	    if (get_one_option(optp->id, optp,
+                               *((my_bool*) value) ?
+                               (char*) "1" : disabled_my_option))
+              return EXIT_ARGUMENT_INVALID;
 	    continue;
 	  }
 	  argument= optend;
 	}
 	else if (optp->arg_type == OPT_ARG &&
-		 (optp->var_type & GET_TYPE_MASK) == GET_BOOL)
+		 (((optp->var_type & GET_TYPE_MASK) == GET_BOOL) ||
+                   (optp->var_type & GET_TYPE_MASK) == GET_ENUM))
 	{
 	  if (optend == disabled_my_option)
 	    *((my_bool*) value)= (my_bool) 0;
@@ -425,7 +464,8 @@ invalid value '%s'",
 		  optp->arg_type == NO_ARG)
 	      {
 		*((my_bool*) optp->value)= (my_bool) 1;
-		get_one_option(optp->id, optp, argument);
+		if (get_one_option(optp->id, optp, argument))
+                  return EXIT_UNSPECIFIED_ERROR;
 		continue;
 	      }
 	      else if (optp->arg_type == REQUIRED_ARG ||
@@ -444,7 +484,8 @@ invalid value '%s'",
                   {
                     if (optp->var_type == GET_BOOL)
                       *((my_bool*) optp->value)= (my_bool) 1;
-                    get_one_option(optp->id, optp, argument);
+                    if (get_one_option(optp->id, optp, argument))
+                      return EXIT_UNSPECIFIED_ERROR;
                     continue;
                   }
 		  /* Check if there are more arguments after this one */
@@ -469,7 +510,8 @@ invalid value '%s'",
                                          my_progname, argument, optp->name);
 		return error;
 	      }
-	      get_one_option(optp->id, optp, argument);
+	      if (get_one_option(optp->id, optp, argument))
+                return EXIT_UNSPECIFIED_ERROR;
 	      break;
 	    }
 	  }
@@ -477,7 +519,7 @@ invalid value '%s'",
 	  {
 	    if (my_getopt_print_errors)
               my_getopt_error_reporter(ERROR_LEVEL,
-                                       "%s: unknown option '-%c'",
+                                       "%s: unknown option '-%c'", 
                                        my_progname, *optend);
 	    return EXIT_UNKNOWN_OPTION;
 	  }
@@ -492,7 +534,8 @@ invalid value '%s'",
                                  my_progname, argument, optp->name);
 	return error;
       }
-      get_one_option(optp->id, optp, argument);
+      if (get_one_option(optp->id, optp, argument))
+        return EXIT_UNSPECIFIED_ERROR;
 
       (*argc)--; /* option handled (short or long), decrease argument count */
     }
@@ -559,14 +602,14 @@ static char *check_struct_option(char *cur_arg, char *key_name)
   Will set the option value to given value
 */
 
-static int setval(const struct my_option *opts, gptr *value, char *argument,
+static int setval(const struct my_option *opts, uchar* *value, char *argument,
 		  my_bool set_maximum_value)
 {
   int err= 0;
 
   if (value && argument)
   {
-    gptr *result_pos= ((set_maximum_value) ?
+    uchar* *result_pos= ((set_maximum_value) ?
 		       opts->u_max_value : value);
 
     if (!result_pos)
@@ -594,6 +637,9 @@ static int setval(const struct my_option *opts, gptr *value, char *argument,
     case GET_ULL:
       *((ulonglong*) result_pos)= getopt_ull(argument, opts, &err);
       break;
+    case GET_DOUBLE:
+      *((double*) result_pos)= getopt_double(argument, opts, &err);
+      break;
     case GET_STR:
       *((char**) result_pos)= argument;
       break;
@@ -603,6 +649,25 @@ static int setval(const struct my_option *opts, gptr *value, char *argument,
       if (!(*((char**) result_pos)= my_strdup(argument, MYF(MY_WME))))
 	return EXIT_OUT_OF_MEMORY;
       break;
+    case GET_ENUM:
+      if (((*(int*)result_pos)=
+             find_type(argument, opts->typelib, 2) - 1) < 0)
+      {
+        /*
+          Accept an integer representation of the enumerated item.
+        */
+        char *endptr;
+        unsigned int arg= (unsigned int) strtol(argument, &endptr, 10);
+        if (*endptr || arg >= opts->typelib->count)
+          return EXIT_ARGUMENT_INVALID;
+        *(int*)result_pos= arg;
+      }
+      break;
+    case GET_SET:
+      *((ulonglong*)result_pos)= find_typeset(argument, opts->typelib, &err);
+      if (err)
+        return EXIT_ARGUMENT_INVALID;
+      break;
     default:    /* dummy default to avoid compiler warnings */
       break;
     }
@@ -694,7 +759,7 @@ my_bool getopt_compare_strings(register const char *s, register const char *t,
   be k|K for kilo, m|M for mega or g|G for giga.
 */
 
-static longlong eval_num_suffix (char *argument, int *error, char *option_name)
+static longlong eval_num_suffix(char *argument, int *error, char *option_name)
 {
   char *endchar;
   longlong num;
@@ -740,7 +805,7 @@ static longlong eval_num_suffix (char *argument, int *error, char *option_name)
 static longlong getopt_ll(char *arg, const struct my_option *optp, int *err)
 {
   longlong num=eval_num_suffix(arg, err, (char*) optp->name);
-  return getopt_ll_limit_value(num, optp);
+  return getopt_ll_limit_value(num, optp, NULL);
 }
 
 /*
@@ -750,11 +815,11 @@ static longlong getopt_ll(char *arg, const struct my_option *optp, int *err)
   Returns "fixed" value.
 */
 
-static longlong getopt_ll_limit_value(longlong num,
-                                      const struct my_option *optp)
+longlong getopt_ll_limit_value(longlong num, const struct my_option *optp,
+                               my_bool *fix)
 {
   longlong old= num;
-  bool trunc= FALSE;
+  my_bool adjusted= FALSE;
   char buf1[255], buf2[255];
   ulonglong block_size= (optp->block_size ? (ulonglong) optp->block_size : 1L);
 
@@ -762,7 +827,29 @@ static longlong getopt_ll_limit_value(longlong num,
       optp->max_value) /* if max value is not set -> no upper limit */
   {
     num= (ulonglong) optp->max_value;
-    trunc= TRUE;
+    adjusted= TRUE;
+  }
+
+  switch ((optp->var_type & GET_TYPE_MASK)) {
+  case GET_INT:
+    if (num > (longlong) INT_MAX)
+    {
+      num= ((longlong) INT_MAX);
+      adjusted= TRUE;
+    }
+    break;
+  case GET_LONG:
+#if SIZEOF_LONG < SIZEOF_LONG_LONG
+    if (num > (longlong) LONG_MAX)
+    {
+      num= ((longlong) LONG_MAX);
+      adjusted= TRUE;
+    }
+#endif
+    break;
+  default:
+    DBUG_ASSERT((optp->var_type & GET_TYPE_MASK) == GET_LL);
+    break;
   }
 
   num= ((num - optp->sub_size) / block_size);
@@ -771,10 +858,13 @@ static longlong getopt_ll_limit_value(longlong num,
   if (num < optp->min_value)
   {
     num= optp->min_value;
-    trunc= TRUE;
+    if (old < optp->min_value)
+      adjusted= TRUE;
   }
 
-  if (trunc)
+  if (fix)
+    *fix= adjusted;
+  else if (adjusted)
     my_getopt_error_reporter(WARNING_LEVEL,
                              "option '%s': signed value %s adjusted to %s",
                              optp->name, llstr(old, buf1), llstr(num, buf2));
@@ -796,9 +886,9 @@ static ulonglong getopt_ull(char *arg, const struct my_option *optp, int *err)
 
 
 ulonglong getopt_ull_limit_value(ulonglong num, const struct my_option *optp,
-                                 bool *fix)
+                                 my_bool *fix)
 {
-  bool adjusted= FALSE;
+  my_bool adjusted= FALSE;
   ulonglong old= num;
   char buf1[255], buf2[255];
 
@@ -840,66 +930,155 @@ ulonglong getopt_ull_limit_value(ulonglong num, const struct my_option *optp,
   if (num < (ulonglong) optp->min_value)
   {
     num= (ulonglong) optp->min_value;
-    adjusted= TRUE;
+    if (old < (ulonglong) optp->min_value)
+      adjusted= TRUE;
   }
 
-  if (adjusted)
+  if (fix)
+    *fix= adjusted;
+  else if (adjusted)
     my_getopt_error_reporter(WARNING_LEVEL,
                              "option '%s': unsigned value %s adjusted to %s",
                              optp->name, ullstr(old, buf1), ullstr(num, buf2));
 
-  if (fix)
-    *fix= adjusted;
-
   return num;
 }
 
 
+/*
+  Get double value withing ranges
+
+  Evaluates and returns the value that user gave as an argument to a variable.
+
+  RETURN
+    decimal value of arg
+
+    In case of an error, prints an error message and sets *err to
+    EXIT_ARGUMENT_INVALID.  Otherwise err is not touched
+*/
+
+static double getopt_double(char *arg, const struct my_option *optp, int *err)
+{
+  double num;
+  int error;
+  char *end= arg + 1000;                     /* Big enough as *arg is \0 terminated */
+  num= my_strtod(arg, &end, &error);
+  if (end[0] != 0 || error)
+  {
+    fprintf(stderr,
+            "%s: ERROR: Invalid decimal value for option '%s'\n",
+            my_progname, optp->name);
+    *err= EXIT_ARGUMENT_INVALID;
+    return 0.0;
+  }
+  if (optp->max_value && num > (double) optp->max_value)
+    num= (double) optp->max_value;
+  return max(num, (double) optp->min_value);
+}
+
+/*
+  Init one value to it's default values
+
+  SYNOPSIS
+    init_one_value()
+    option		Option to initialize
+    value		Pointer to variable
+*/
+
+static void init_one_value(const struct my_option *option, uchar* *variable,
+			   longlong value)
+{
+  DBUG_ENTER("init_one_value");
+  switch ((option->var_type & GET_TYPE_MASK)) {
+  case GET_BOOL:
+    *((my_bool*) variable)= (my_bool) value;
+    break;
+  case GET_INT:
+    *((int*) variable)= (int) getopt_ll_limit_value((int) value, option, NULL);
+    break;
+  case GET_ENUM:
+    *((uint*) variable)= (uint) value;
+    break;
+  case GET_UINT:
+    *((uint*) variable)= (uint) getopt_ull_limit_value((uint) value, option, NULL);
+    break;
+  case GET_LONG:
+    *((long*) variable)= (long) getopt_ll_limit_value((long) value, option, NULL);
+    break;
+  case GET_ULONG:
+    *((ulong*) variable)= (ulong) getopt_ull_limit_value((ulong) value, option, NULL);
+    break;
+  case GET_LL:
+    *((longlong*) variable)= (longlong) getopt_ll_limit_value((longlong) value, option, NULL);
+    break;
+  case GET_ULL:
+  case GET_SET:
+    *((ulonglong*) variable)= (ulonglong) getopt_ull_limit_value((ulonglong) value, option, NULL);
+    break;
+  case GET_DOUBLE:
+    *((double*) variable)=  (double) value;
+    break;
+  case GET_STR:
+    /*
+      Do not clear variable value if it has no default value.
+      The default value may already be set.
+      NOTE: To avoid compiler warnings, we first cast longlong to intptr,
+      so that the value has the same size as a pointer.
+    */
+    if ((char*) (intptr) value)
+      *((char**) variable)= (char*) (intptr) value;
+    break;
+  case GET_STR_ALLOC:
+    /*
+      Do not clear variable value if it has no default value.
+      The default value may already be set.
+      NOTE: To avoid compiler warnings, we first cast longlong to intptr,
+      so that the value has the same size as a pointer.
+    */
+    if ((char*) (intptr) value)
+    {
+      my_free((*(char**) variable), MYF(MY_ALLOW_ZERO_PTR));
+      *((char**) variable)= my_strdup((char*) (intptr) value, MYF(MY_WME));
+    }
+    break;
+  default: /* dummy default to avoid compiler warnings */
+    break;
+  }
+  DBUG_VOID_RETURN;
+}
+
+
 /*
   Init one value to it's default values
 
   SYNOPSIS
     init_one_value()
-    optp               Option to initialize
-    value              Pointer to variable
+    option		Option to initialize
+    value		Pointer to variable
 */
 
-static void init_one_value(const struct my_option *optp, gptr *variable,
-			   longlong value)
+static void fini_one_value(const struct my_option *option, uchar* *variable,
+			   longlong value __attribute__ ((unused)))
 {
-  DBUG_ENTER("init_one_value");
-
-  switch ((optp->var_type & GET_TYPE_MASK)) {
-  case GET_BOOL:
-    *((my_bool*) variable)= (my_bool) value;
-    break;
-  case GET_INT:
-    *((int*) variable)= (int) getopt_ll_limit_value(value, optp);
-    break;
-  case GET_UINT:
-    *((uint*) variable)= (uint) getopt_ull_limit_value(value, optp, NULL);
-    break;
-  case GET_LONG:
-    *((long*) variable)= (long) getopt_ll_limit_value(value, optp);
-    break;
-  case GET_ULONG:
-    *((ulong*) variable)= (ulong) getopt_ull_limit_value(value, optp, NULL);
-    break;
-  case GET_LL:
-    *((longlong*) variable)= (longlong) getopt_ll_limit_value(value, optp);
-    break;
-  case GET_ULL:
-    *((ulonglong*) variable)=  (ulonglong) getopt_ull_limit_value(value, optp,
-                                                                  NULL);
+  DBUG_ENTER("fini_one_value");
+  switch ((option->var_type & GET_TYPE_MASK)) {
+  case GET_STR_ALLOC:
+    my_free((*(char**) variable), MYF(MY_ALLOW_ZERO_PTR));
+    *((char**) variable)= NULL;
     break;
   default: /* dummy default to avoid compiler warnings */
     break;
   }
-
   DBUG_VOID_RETURN;
 }
 
 
+void my_cleanup_options(const struct my_option *options)
+{
+  init_variables(options, fini_one_value);
+}
+
+
 /* 
   initialize all variables to their default values
 
@@ -913,12 +1092,13 @@ static void init_one_value(const struct my_option *optp, gptr *variable,
     for a value and initialize.
 */
 
-static void init_variables(const struct my_option *options)
+static void init_variables(const struct my_option *options,
+                           init_func_p init_one_value)
 {
   DBUG_ENTER("init_variables");
   for (; options->name; options++)
   {
-    gptr *variable;
+    uchar* *variable;
     DBUG_PRINT("options", ("name: '%s'", options->name));
     /*
       We must set u_max_value first as for some variables
@@ -930,7 +1110,7 @@ static void init_variables(const struct my_option *options)
     if (options->value)
       init_one_value(options, options->value, options->def_value);
     if (options->var_type & GET_ASK_ADDR &&
-	(variable= (*getopt_get_addr)("", 0, options)))
+	(variable= (*getopt_get_addr)("", 0, options, 0)))
       init_one_value(options, variable, options->def_value);
   }
   DBUG_VOID_RETURN;
@@ -1023,7 +1203,8 @@ void my_print_help(const struct my_option *options)
 
 void my_print_variables(const struct my_option *options)
 {
-  uint name_space= 34, length;
+  uint name_space= 34, length, nr;
+  ulonglong bit, llvalue;
   char buff[255];
   const struct my_option *optp;
 
@@ -1032,8 +1213,8 @@ void my_print_variables(const struct my_option *options)
   printf("--------------------------------- -----------------------------\n");
   for (optp= options; optp->id; optp++)
   {
-    gptr *value= (optp->var_type & GET_ASK_ADDR ?
-		  (*getopt_get_addr)("", 0, optp) : optp->value);
+    uchar* *value= (optp->var_type & GET_ASK_ADDR ?
+		  (*getopt_get_addr)("", 0, optp, 0) : optp->value);
     if (value)
     {
       printf("%s ", optp->name);
@@ -1041,6 +1222,21 @@ void my_print_variables(const struct my_option *options)
       for (; length < name_space; length++)
 	putchar(' ');
       switch ((optp->var_type & GET_TYPE_MASK)) {
+      case GET_SET:
+        if (!(llvalue= *(ulonglong*) value))
+	  printf("%s\n", "(No default value)");
+	else
+        for (nr= 0, bit= 1; llvalue && nr < optp->typelib->count; nr++, bit<<=1)
+	{
+	  if (!(bit & llvalue))
+	    continue;
+	  llvalue&= ~bit;
+	  printf( llvalue ? "%s," : "%s\n", get_type(optp->typelib, nr));
+	}
+	break;
+      case GET_ENUM:
+        printf("%s\n", get_type(optp->typelib, *(uint*) value));
+	break;
       case GET_STR:
       case GET_STR_ALLOC:                    /* fall through */
 	printf("%s\n", *((char**) value) ? *((char**) value) :
@@ -1068,6 +1264,9 @@ void my_print_variables(const struct my_option *options)
 	longlong2str(*((ulonglong*) value), buff, 10);
 	printf("%s\n", buff);
 	break;
+      case GET_DOUBLE:
+	printf("%g\n", *(double*) value);
+	break;
       default:
 	printf("(Disabled)\n");
 	break;
